@@ -258,6 +258,7 @@ void TocHandler::openForRead() const {
     }
 #endif
     SYSCALL2((fd_ = ::open( tocPath_.localPath(), iomode )), tocPath_ );
+    eckit::Length tocSize = tocPath_.size();
 
     // The masked subtocs and indexes could be updated each time, so reset this.
     enumeratedMaskedEntries_ = false;
@@ -269,16 +270,36 @@ void TocHandler::openForRead() const {
         AutoClose closer1(toc);
         fd_ = -1;
 
+
         bool grow = true;
-        cachedToc_.reset( new eckit::MemoryHandle(tocPath().size(), grow) );
+        cachedToc_.reset( new eckit::MemoryHandle(tocSize, grow) );
 
         long buffersize = 4*1024*1024;
-        toc.copyTo(*cachedToc_, buffersize);
-
+        toc.copyTo(*cachedToc_, buffersize, tocSize, tocReadStats_);
         cachedToc_->openForRead();
     }
 }
 
+void TocHandler::dumpTocCache() const {
+    eckit::Offset offset = cachedToc_->position();
+    cachedToc_->seek(0);
+
+    eckit::PathName tocDumpFile("dump_of_"+tocPath_.baseName());
+    eckit::FileHandle dump(eckit::PathName::unique(tocDumpFile));
+    cachedToc_->copyTo(dump);
+
+    std::ostringstream ss;
+    ss << tocPath_.baseName() << " read in " << tocReadStats_.size() << " step" << ((tocReadStats_.size()>1)?"s":"") << std::endl;
+    double time;
+    eckit::Length len;
+    while (tocReadStats_.next(time, len)) {
+        ss << "  step duration: " << (time*1000) << " ms, size: " << len << " bytes"<< std::endl;
+    }
+    Log::error() << ss.str();
+
+    cachedToc_->seek(offset);
+}
+
 void TocHandler::append(TocRecord &r, size_t payloadSize ) {
 
     ASSERT(fd_ != -1);
@@ -432,14 +453,24 @@ bool TocHandler::readNextInternal(TocRecord& r) const {
 
     CachedFDProxy proxy(tocPath_, fd_, cachedToc_);
 
-    long len = proxy.read(&r, sizeof(TocRecord::Header));
-    if (len == 0) {
-        return false;
+    try {
+        long len = proxy.read(&r, sizeof(TocRecord::Header));
+        if (len == 0) {
+            return false;
+        }
+        ASSERT(len == sizeof(TocRecord::Header));
+    } catch(...) {
+        dumpTocCache();
+        throw;
     }
-    ASSERT(len == sizeof(TocRecord::Header));
 
-    len = proxy.read(&r.payload_, r.header_.size_ - sizeof(TocRecord::Header));
-    ASSERT(size_t(len) == r.header_.size_ - sizeof(TocRecord::Header));
+    try {
+        long len = proxy.read(&r.payload_, r.header_.size_ - sizeof(TocRecord::Header));
+        ASSERT(size_t(len) == r.header_.size_ - sizeof(TocRecord::Header));
+    } catch(...) {
+        dumpTocCache();
+        throw;
+    }
 
     LibFdb5::instance().serialisationVersion().check(r.header_.version_);
 
